// Once.h
// (c) Tony Van Eerd Sept 30, 2009
//
// This version of call_once
// attempts to avoid creating a synchronization primitive (ie Event under Windows)
// unless it is actually necessary - ie event creation is delayed until actual contention exists.
// This is in stark contrast to most implementations, many of which:
//  - are broken (ie priority inversion problems)
//  - tend to create sync-objects on the first thread in (ie thus guaranteeing one sync-object for each once_flag)
//  - create heavier objects (ie named mutexes or named events - which are global across all processes).
// see, for example, Boost threads, and most linux implementations.
// This is particularly important when you see code with 100s or 1000s of singletons (yes I've seen that).
// See also DCLP article by Scott Myers and Andre Alexandrescu.
//

//
// Method(s)/Notes/Tricks:
//
// - track *all* 'stages' of progress using the single atomic once_flag - ie avoid having multiple variables for storing state
// - multiple waiting threads are 'chained' together in a node based chain (with the once_flag used as head of the chain)
// - to avoid any memory allocations, each node is stored on the stack of its thread
// - the chain is carefully managed to ensure it doesn't disappear prematurely while some waiting threads are still on it
// - handles case where to-be-called-once function throws an exception
// - optionally allows recursion, as odd as that is
//

#include <windows.h>  // Interlocked functions, Event handle functions
#include <assert.h>
#include <cstddef>  // size_t

namespace detail
{
	template <bool b>
	struct static_assert_test
	{
	};

	template <>
	struct static_assert_test<true>
	{
		typedef bool type;
		enum { value = true };
	};
}

#define static_assert(x)  typedef detail::static_assert_test<!!(x)>::type static_assert_typedef_##__LINE__;


namespace atomic
{
	typedef int int_t; // whatever type is atomic on the platform
	typedef unsigned int uint_t;   // whatever type is atomic on the platform
	typedef intptr_t intptr_t;



#pragma warning( push )
#pragma warning( disable : 4311 4312 ) // casting ptr to int or vice versa


	template <std::size_t size>
	struct by_size
	{
	};
	template <> struct by_size<4>
	{
		template <typename T1, typename T2, typename T3>
		static T1 CAS(T1 volatile & target, T2 compare, T3 newval)
		{
			static_assert(sizeof(T1) == sizeof(LONG));
			return (T1)InterlockedCompareExchange((LONG volatile *)&target, (LONG)newval, (LONG)compare);
		}

		template <typename T1, typename T2>
		static T1 write(T1 volatile & target, T2 newval)
		{
			static_assert(sizeof(T1) == sizeof(LONG));
			return (T1)InterlockedExchange/*Release*/((LONG volatile *)&target, (LONG)newval);
		}
		template <typename T1>
		static T1 read(T1 volatile & target)
		{
			static_assert(sizeof(T1) == sizeof(LONG));
			return (T1)InterlockedCompareExchangeAcquire((LONG volatile *)&target, 0, 0);
		}

		template <typename T1>
		static T1 increment(T1 volatile & target)
		{
			static_assert(sizeof(T1) == sizeof(LONG));
			return (T1)InterlockedIncrement((LONG volatile *)&target);
		}
		template <typename T1>
		static T1 decrement(T1 volatile & target)
		{
			static_assert(sizeof(T1) == sizeof(LONG));
			return (T1)InterlockedDecrement((LONG volatile *)&target);
		}
	};
	template <> struct by_size<8>
	{
		template <typename T1, typename T2, typename T3>
		static T1 CAS(T1 volatile & target, T2 compare, T3 newval)
		{
			static_assert(sizeof(T1) == sizeof(LONGLONG));
			return (T1)InterlockedCompareExchange64((LONGLONG volatile *)&target, (LONGLONG)newval, (LONGLONG)compare);
		}

		template <typename T1, typename T2>
		static T1 write(T1 volatile & target, T2 newval)
		{
			static_assert(sizeof(T1) == sizeof(LONGLONG));
			return (T1)InterlockedExchange64/*Release*/((LONGLONG volatile *)&target, (LONGLONG)newval);
		}
		template <typename T1>
		static T1 read(T1 volatile & target)
		{
			static_assert(sizeof(T1) == sizeof(LONGLONG));
			return (T1)InterlockedCompareExchangeAcquire64((LONGLONG volatile *)&target, 0, 0);
		}

		template <typename T1>
		static T1 increment(T1 volatile & target)
		{
			static_assert(sizeof(T1) == sizeof(LONGLONG));
			return (T1)InterlockedIncrement64((LONGLONG volatile *)&target);
		}
		template <typename T1>
		static T1 decrement(T1 volatile & target)
		{
			static_assert(sizeof(T1) == sizeof(LONGLONG));
			return (T1)InterlockedDecrement64((LONGLONG volatile *)&target);
		}
	};
#pragma warning( pop ) 


	template <typename T1, typename T2, typename T3>
	T1 CAS(T1 volatile & target, T2 compare, T3 newval)
	{
		static_assert(sizeof(T2) <= sizeof(T1));
		static_assert(sizeof(T3) <= sizeof(T1));

		return by_size<sizeof(T1)>::CAS(target, compare, newval);
	}

	template <typename T1, typename T2>
	T1 write(T1 volatile & target, T2 newval)
	{
		static_assert(sizeof(T2) <= sizeof(T1));

		return by_size<sizeof(T1)>::write(target, newval);
	}

	template <typename T1>
	T1 read(T1 volatile & target)
	{
		return by_size<sizeof(T1)>::read(target);
	}

	template <typename T1>
	T1 increment(T1 volatile & target)
	{
		return by_size<sizeof(T1)>::increment(target);
	}
	template <typename T1>
	T1 decrement(T1 volatile & target)
	{
		return by_size<sizeof(T1)>::decrement(target);
	}
}


//typedef atomic::int_t once_flag;

struct once_flag
{
	typedef atomic::intptr_t atomic_flag;
	atomic_flag flag;
	DWORD thread_id;
};


namespace detail
{

template <typename Intish>
Intish high_half(Intish val)
{
	return val >> (CHAR_BIT * sizeof(val) / 2);  // if Intish is signed, then this does sign-extension, so negatives stay negative. For better or worse.
}

template <typename Intish>
Intish low_half(Intish val)
{
	return high_half(val << (CHAR_BIT * sizeof(val) / 2));
}

template <typename Intish>
Intish make_whole(Intish low, Intish high)
{
	// low better be only on low half
	// could truncate or assert, etc.
	return low | (high << (CHAR_BIT * sizeof(Intish) / 2));
}

struct vcall
{
	virtual void call()
	{
	}
};

template <typename Callable>
struct callable : vcall
{
	Callable func;
	callable(Callable function) : func(function)
	{
	}
	virtual void call()
	{
		func();
	}
};




enum { once_done = 1 };

struct once_node
{
	once_node *next;

	HANDLE done;
	HANDLE cleanup;
	atomic::uint_t count;
	DWORD thread_id;

	once_node()
		:
		next(0),
		done(0),
		cleanup(0),
		count(0)
	{
	}

	~once_node()
	{
		if (done)
			CloseHandle(done);
		if (cleanup)
			CloseHandle(cleanup);
	}
};

once_flag::atomic_flag add_winner_bit(once_node * addr)
{
	assert(((once_flag::atomic_flag)addr & 0x3) == 0); // must be aligned for CAS, also for our algorithm
	return (once_flag::atomic_flag)addr | (once_flag::atomic_flag)0x1;
}
once_node * remove_winner_bit(once_flag::atomic_flag flag)
{
	return (once_node *)(flag & ~(once_flag::atomic_flag)0x1);
}
bool has_winner_bit(once_flag::atomic_flag flag)
{
	return !!(flag & (once_flag::atomic_flag)0x1);
}
bool has_winner_bit(once_node *node)
{
	return has_winner_bit((once_flag::atomic_flag)node);
}


void post_winner_cleanup(once_flag & flag, bool success)
{
	// called by 'winner' thread after calling func.
	// called whether func was successful or NOT (ie even if func throws)

	// first, set flag to success/fail
	// if fail, then some other thread (if any are waiting) will need to retry
	// (of course it will probably fail/throw for the same reasons that this thread did,
	//  but at least we tried :-)
	once_flag::atomic_flag oldflag = atomic::write(flag.flag, success ? once_done : 0);

	// oldflag is a chain of once_nodes.
	// Now that we have swapped it out from the once_flag, it is no longer visible, so no threads can add to it!
	// (This makes things easier! Not easy, but easier :-)

	// if first entry in chain is winner (us),
	// then there are no other threads to consider...
	if (has_winner_bit(oldflag))
	{
		// no other threads waiting, nothing to do.
		return;
	}

	// ... else there is a chain of waiting threads:
	once_node * chain = (once_node *)oldflag;
	atomic::uint_t true_count = 1;

	// find us at end of chain
	while (!has_winner_bit(chain->next))
	{
		chain = chain->next;
		true_count++;
	}

	// we are now at the end of the chain
	// just *before* us/winner
	// ie at the first waiter
	once_node * endwaiter = (once_node *)chain;

	if (true_count > 1)
	{
		// multiple waiters. endwaiter needs to wait for the rest of them!
		// Ensure that there is a cleanup event for the endwaiter to wait on,
		// in case none of the additional waiters has managed to set that up yet.
		if (!endwaiter->cleanup) // relaxed read
		{
			HANDLE cleanup = CreateEvent(0, false, false, 0);
			HANDLE oldcleanup = atomic::CAS(endwaiter->cleanup, 0, cleanup);
			if (oldcleanup)
				CloseHandle(cleanup); // someone beat us to it. fine.
		}
	}

	atomic::uint_t curr_count = atomic::read(endwaiter->count);
	atomic::uint_t prev_count = curr_count;

	do
	{
		// true_count is how many waiters we counted in the list
		// curr_count is how many waiters have walked the chain and annouced their existence by incrementing endwaiter->count
		if (true_count == curr_count)
		{
			// all waiters are *at least* 'on' the end node, although they may not yet be waiting on the done event.
			// done event is manual reset, so once signaled, it will stay open, so every thread will wake up, even if they haven't even started waiting yet
			// (ie those threads will pass right through the wait function)
			SetEvent(endwaiter->done);
			break;
		}
		else
		{
			// some waiter is still walking the chain.
			// let them know that the chain has been disconnected from the once_flag
			// so there can be no new waiters.
			// From here (this current thread) we can know the real length of the chain,
			// so expose that to the 'slow' chain walkers, and let them trigger the done event
			// when the last slow walker gets to the end of the chain.

			// we stick the true count in the high-half of endwaiter->count, current running count in low half
			curr_count = atomic::CAS(endwaiter->count, curr_count, make_whole(curr_count, true_count));

			// if the CAS fails (ie if end->count moved since we last saw it),
			// then the do...while() will loop around again, and maybe we'll set the event next time, or get a successful CAS. (eventually)
		}
	}
	while (curr_count != prev_count); // did things change while we were making our attempt? or did we successfully CAS
}


void call_once(once_flag & flag, vcall & func, bool to_completion, bool allow_recursion)
{
	once_node node;

	static_assert(sizeof(flag.flag) == sizeof(&node));

	// try to be first node in (ie winner)
	// if we fail, we still get an atomic read,
	// which we need anyway
	// must have *acquire* memory ordering to ensure that all data protected by the once_flag is visible (particularly in the done case)
	once_flag::atomic_flag oldflag = atomic::CAS(flag.flag, 0, add_winner_bit(&node));

	if (oldflag == once_done)
	{
		return;
	}

	bool check_for_recursion = allow_recursion;
	DWORD thread_id;
#ifndef NDEBUG
	check_for_recursion = true;
#endif

	if (check_for_recursion)
	{
		thread_id = GetCurrentThreadId();
		if (allow_recursion)
		{
			// don't need memory-ordered reads at all for this comparison.
			// if we are in the same thread (recursing), then obviously we don't need a barrier
			// if in another thread (or first thread in), then either we see old value of flag.thread_id (0)
			// or new value, something != our thread_id
			if (flag.thread_id == thread_id)
			{
				// note that there is no try-catch for the recursive call -
				// only catch at the outermost/first call
				// (see case 0 in switch statement below)
				func.call();

				// don't call post_winner_cleanup() here!
				// just un-recurse and let it be called at the top/beginning of the recursion
				// (inside the switch case 0: below).
				// Otherwise if we set flag to once_done here, we lose the chain of waiters.
				// (Actually it would probably still work, but only because, by coincidence, has_winner_bit(once_done) == true,
				//  and thus the post_winner_cleanup() at the top of the recursion cleanly exits early.)
				// More importantly, func.call() could still throw during the unravelling of the recursion,
				// so we are NOT *done* yet!
				// (also func.call() could also catch its own exception during the un-recurse!)
				// ie not done until we get back to the top of the recursion.  So just get out of here.
				return;
			}
		}
		else
		{
			assert(flag.thread_id != thread_id);
		}
	}


	// note that we loop around until done.
	// This means a thread can come in, attempt the call, and fail,
	// all while other threads are waiting.
	// For the 'to_completion' case, failure means not done, so:
	// on failure (exception), the 'winner' (loser?) leaves with the exception (doesn't get to try again)
	//  - ie sort of as if the call was never made (at least not completed).
	// The waiters (if any) will awaken, loop around, and one of them will 'win' this time, and try again.
	// (Of course if the first thread failed, they probably all will, but who knows.
	//  Worst case, they each get an exception thrown in their own thread, so they have some idea what happened.)
	// (And note that this all works because of how we swap out the chain of waiter nodes, and thus we can reuse the once_flag for a new chain!)

	while (oldflag != once_done)
	{
		if (oldflag == 0) // firsttime!
		{
			// we have a winner! we're first in, go for it

			if (check_for_recursion)
				flag.thread_id = thread_id;  // mark it so other calls in can check

			try
			{
				func.call();
			}
			catch (...)
			{
				// we have some cleanup to do before rethrow
				post_winner_cleanup(flag, !to_completion); // called, but didn't complete
				throw;
			}

			post_winner_cleanup(flag, true);  // done, and complete
			return;
		}
		else
		{
			// flag/oldflag is already a pointer to a once_node.
			// get in line!
			if (has_winner_bit(oldflag))
			{
				node.count = 1;  // preset this in case we are the first waiter, otherwise it can be anything
				// oldflag is node from first thread in,
				// so we *might be* the second thread in, if we get in line soon enough.
				// All waiters will then wait on our done event.
				// NOTE that we need to have the event created *before* we get in line,
				// so that it is already there for additional waiters if they are right behind us.
				node.done = CreateEvent(0, true, false, 0);  // manual reset, initial state is non-signaled
			}

			// this is just like the classic push onto a lockfree stack
			do
			{
				node.next = (once_node *)oldflag;
				oldflag = atomic::CAS(flag.flag, oldflag, &node); // if it hasn't changed yet, swap us into the line
			}
			while (oldflag != (once_flag::atomic_flag)node.next); // ie if another thread pushed while we were trying to

			if (has_winner_bit(oldflag))
			{
				// we made it as the second thread in.!

				// we now just need to wait for the done event to be signaled (by the winner thread - probably)
				// note also that all other threads coming in behind us will wait on this as well
				WaitForSingleObject(node.done, INFINITE);

				// OK, signalled.  So now we are back...

				// if more than just this thread was waiting,
				// then all the other waiters were/are reading this node (on our stack!) and/or waiting on *our* event
				// so we can't leave (trashing our stack) nor cleanup the done event (while other threads are just about to wait on it).
				// Instead, rest assured that those other threads will let us know when they are done:

				// we keep track of number of waiters left (which included us):
				atomic::uint_t count = atomic::decrement(node.count);
				if (count > 0)
				{
					// other threads still waiting on done!
					// we can't leave (nor clean up handles) until they are all finished waiting.
					assert(node.cleanup != NULL); // a later waiter should have set this up for us
					WaitForSingleObject(node.cleanup, INFINITE);
				}
			}
			else
			{
				// we are not the first, nor the second thread in

				if (node.done)
				{
					// we had created this thinking we might become the first waiter, but we lost the race :-(
					CloseHandle(node.done);
					node.done = 0;
				}

				// now, walk the chain to the end,
				// wait on the event you find there
				// Note that by this time, the call_once 'winner' thread might be done,
				// and already be trying to walk the chain to wake us up,
				// even though we haven't got there yet.
				// Luckily it is careful not to allow anyone in the chain to leave before we get to the end.
				once_node *endwaiter = node.next;
				while (!has_winner_bit(endwaiter->next))
					endwaiter = endwaiter->next;
				// now at the one-before-winner, or end-waiter

				// since we are the 2nd or nth thread waiting,
				// we need to tell the first-waiter (endwaiter, the owner of the end of the chain)
				// that it can't leave until every other waiter leaves.
				// We do this with the endwaiter->cleanup event handle.
				// ** NOTE that we set up the cleanup event BEFORE incrementing endwaiter->count (farther below),
				//    so that endwaiter can't wake up on 'done' and find endwaiter->cleanup missing. **

				// since we don't know whether we are 2nd or nth,
				// we attempt to Create the cleanup event and race to set end->cleanup
				// if we don't get there first, that's fine, just throw ours away and use what is there.
				{
					HANDLE cleanup = atomic::read(endwaiter->cleanup);    // this is atomic so we don't get a very stale read of our view of another thread's leftover stack
					if (!cleanup)                                         // we could avoid the first atomic::read, and just always attempt the CAS,
						cleanup = CreateEvent(0, false, false, 0);        // but it is probably best to avoid the CreateEvent call at the expence of an additional atomic op
					HANDLE prev_cleanup = atomic::CAS(endwaiter->cleanup, 0, cleanup);
					if (prev_cleanup != 0) // if we didn't get our handle in there
						CloseHandle(cleanup);
				}

				// the 'winning' thread (that did the func call), when done
				// 1. removes the chain from the once_flag, so it is no longer visible,
				//    and cannot be added to.
				// 2. walks the chain, counting the number of 'waiters' in the chain.
				//
				// if it finds that count is the same as the number of nodes in the chain,
				// it knows everyone is on the end of the chain (maybe not yet waiting, but on the end)
				// so all OTHER chained threads except the last one can be awakened and let go (just go, no need to 'pop' from the list);
				// only the end node needs to stick around to clean up (the end node struct is in the STACK space of
				// the thread, so the thread can't leave until all other threads are done with the node).
				//
				// Now, even worse, if a 'waiter' is still walking the chain when 'winner' finishes,
				// *none* of the threads in the chain can leave, until all walkers get to the end.
				// We track this via endwaiter->count.  Walker/waiters increment the low-half of count,
				// the 'winner' thread sets the true count in the high-half of the count.
				// This way either side can see the true state.

				atomic::uint_t count = atomic::increment(endwaiter->count); // we only increment 'bottom-half' of count (ie assume number of waiters doesn't overflow bottom half of an int)

				if (high_half(count) == low_half(count))
				{
					// 'winner' thread stopped by and saw how slow we are/were;
					// it set the true_count in the high-half (otherwise 0) and then impatiently left - without signalling done!
					// Now that we've finally incremented the low-half to catch up to the high-half,
					// that means we must be the last walker/waiter, and it is up to us to flag the done event

					// we can write end->count NON-atomically because there are no other threads left to touch it
					// and the SetEvent() does the proper memory barriers for us (we assume - else most of Windows would be broken)

					endwaiter->count = low_half(count) - 1; // we don't need to wait ourselves, so don't count this thread
					SetEvent(endwaiter->done);  // do the job that 'winner' thread left for us!
				}
				else
				{
					// well, we've added ourselves to the count,
					// looks like the winner thread hasn't been here (or there are more slow-pokes yet to come);
					// nothing left to do but wait on the done event
					WaitForSingleObject(endwaiter->done, INFINITE);

					// and now we are back.
					// decr the count, see if we are last ones left,
					// if so, turn out the lights.  Or, actually, notify the waiting end-node thread that it can now turn out the lights (ie cleanup)
					count = atomic::decrement(endwaiter->count);
					if (count == 0)
					{
						// we were the last one waiting on done.
						// But end-node is still waiting to clean up!
						SetEvent(endwaiter->cleanup);
					}
				}
			}
		}
	}
}

}


//
// calls any callable object/function *once*, even in the presence of threads.
// no thread will return unless
//  - it throws
//  - or one thread made the call
//
// 'to_completion': ie what to do if 'func' throws - do you consider the func 'done'?
// well it was _called_, but didn't complete.
// Set to_completion to true if you need func to succeed (ie not throw) _once_.
// Set to_completion to false if you only need func to be _called_ once (even if it throws).
// Even though the function name is "call_once" (not "complete_once"), the default is to
// require 'completion'.  This is mostly for historic/compatibility reasons
//  - ie pthreads_callonce() behaves similarly (with respect to thread cancelation, not throwing)
//
// 'allow_recursion': normally call_once + recursion doesn't make much sense, and is an error;
// but since recursion *is* a normal and useful programming technique, and it isn't that hard for us to support, why not?
// Of course, it defaults to false to to match standard behaviour.
// if (allow_recursion == false) then recursion asserts in debug, hangs in release.
//

template <typename Callable>
void call_once(once_flag & flag, Callable func, bool to_completion = true, bool allow_recursion = false)
{
	struct detail::callable<Callable> call(func);

	detail::call_once(flag, call, to_completion, allow_recursion);
}

